#include "license.h"
#include <lua.h>
#include <lualib.h>
#include <lauxlib.h>

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>

#include "license_logging.h"
#include "license_objects.h"
#include "license_reader.h"
#include "license_lua_userdata.h"
#include "license_query.h"
#include "xalloc.h"


static LicenseStatus op5lic_legacy_fallback(const char *file, lua_State *L, struct op5lic_error_t *err) {
	LicenseStatus status;
	char * legacy_license_path;
	char * p;
	status = OP5LIC_STATUS_OK;
	p = strrchr(file, '.');
	if (p == NULL || strcmp(p, ".lic") != 0) {
		log_err("Failed to determine fallback license name for %s", file);
		op5lic_error_set(err, OP5LIC_STATUS_NOT_EXIST, NULL);
		return OP5LIC_STATUS_NOT_EXIST;
	}
	legacy_license_path = xstrdup(file);
	strcpy(legacy_license_path + strlen(file) - 4, ".xml");
	log_debug("Legacy license path: %s", legacy_license_path);
	status = op5lic_reader_load_v1_v2_fallback(legacy_license_path, L, err);
	xfree(legacy_license_path);
	return status;
}
struct op5lic_t *op5lic_create(const char *file, struct op5lic_error_t *err) {
	struct op5lic_t *lic;
	struct op5lic_reader_t *reader;
	struct op5lic_error_t *auto_err = NULL;
	LicenseStatus status;

	lic = xcalloc(1, sizeof(struct op5lic_t));
	lic->luaState = luaL_newstate();
	/* Don't open all the standard library files, we just need a really small subset */
	/*
	   lua_pushcfunction(lic->luaState, luaopen_string);
	   lua_call(lic->luaState, 0, 0);
	   lua_pushcfunction(lic->luaState, luaopen_package);
	   lua_call(lic->luaState, 0, 0);
	   lua_pushcfunction(lic->luaState, luaopen_io);
	   lua_call(lic->luaState, 0, 0);
	   lua_pushcfunction(lic->luaState, luaopen_os);
	   lua_call(lic->luaState, 0, 0);
	   lua_pushcfunction(lic->luaState, luaopen_debug);
	   lua_call(lic->luaState, 0, 0);

	*/
	lua_pushcfunction(lic->luaState, luaopen_base);
	lua_call(lic->luaState, 0, 0);
	/* Disable unwanted base functions. We need some of them, though - ipairs
	 * in particular. */
	lua_getglobal(lic->luaState, "_G"); /*_G because globals */
	lua_pushnil(lic->luaState);
	lua_setfield(lic->luaState, -2, "loadstring");
	lua_pushnil(lic->luaState);
	lua_setfield(lic->luaState, -2, "load");
	lua_pushnil(lic->luaState);
	lua_setfield(lic->luaState, -2, "loadfile");
	lua_pushnil(lic->luaState);
	lua_setfield(lic->luaState, -2, "collectgarbage");
	lua_pushnil(lic->luaState);
	lua_setfield(lic->luaState, -2, "dofile");
	lua_pop(lic->luaState, 1);

	lua_pushcfunction(lic->luaState, luaopen_table);
	lua_call(lic->luaState, 0, 0);
	lua_pushcfunction(lic->luaState, luaopen_math);
	lua_call(lic->luaState, 0, 0);

	op5lic_query_luaregister_libfunctions(lic->luaState);
	op5lic_query_luaregister_userdata_response(lic->luaState);
	op5lic_query_luaregister_userdata_env(lic->luaState);

	/*
	 * regardless of whether the user is interested in the error or not,
	 * we need it.
	 */
	if (err == NULL) {
		auto_err = op5lic_error_create();
		err = auto_err;
	}

	/* Load the license file */
	reader = op5lic_reader_create(file, err);
	status = err->status;
	if (reader == NULL) {
		log_err("Failed to create license reader: %s", op5lic_strerror(err));
		if (status == OP5LIC_STATUS_NOT_EXIST) {
			/* The v3 license did not exist, see if we can find a v1/v2 license
			 * and use that instead
			 */

			/* reset err */
			op5lic_error_set(err, OP5LIC_STATUS_OK, NULL);
			status = op5lic_legacy_fallback(file, lic->luaState, err);
			if (status != OP5LIC_STATUS_OK) {
				log_err("Failed to create fallback license: %s", op5lic_strerror(err));
				goto err_out;
			}
		}
		else if (status != OP5LIC_STATUS_OK) {
			log_err("Failed to create license: %s", op5lic_strerror(err));
			goto err_out;
		}
	}
	else if ((status = op5lic_reader_load_lua(reader, lic->luaState, err)) != OP5LIC_STATUS_OK) {
		log_err("Failed to load lua: %s", op5lic_strerror(err));
		goto err_out;

	}

err_out:
	if (status != OP5LIC_STATUS_OK) {
		op5lic_destroy(lic);
		lic = NULL;
	}

	if (auto_err)
		op5lic_error_destroy(auto_err);

	if (reader)
		op5lic_reader_destroy(reader);

	return lic;
}

void op5lic_destroy(struct op5lic_t *lic) {
	if (lic == NULL) {
		return;
	}
	lua_close(lic->luaState);
	xfree(lic);
}
